from tqdm import tqdm

from easymlops.core import *
import pandas as pd
import copy
import numpy as np

dataframe_type = pd.DataFrame
series_type = pd.Series
dict_type = dict
GlobalNullValues = [np.nan, np.inf, None, "NULL", "NaN", "Nan", "None", "NONE", "none", "nan", "np.nan",
                    "null", "", " ", "inf", "np.inf", "Null"]
GlobalExtremeValues = [np.inf, 0.0, 0, -1, -1.0, 1.0, 1, -1e-7, 1e-7, np.iinfo(np.int64).min,
                       np.iinfo(np.int64).max, np.finfo(np.float64).min,
                       np.finfo(np.float64).max, "", "null", None, "1.0", "0.0", "-1.0", "-1", "NaN", "None"]


class TablePipeObjectBase(PipeObjectBase):
    """
    Table型模型的PipeBase
    """

    def __init__(self, name=None, transform_check_max_number_error=1e-5, skip_check_transform_type=False,
                 skip_check_transform_value=False, leak_check_transform_type=True, leak_check_transform_value=True,
                 copy_data=False, prefix=None, **kwargs):
        """

        :param name: 模块名称，如果为空默认为self.__class__
        :param transform_check_max_number_error: 在check_transform_function时允许的最大数值误差
        :param skip_check_transform_type: 在check_transform_function时是否跳过类型检测(针对一些特殊数据类型，比如稀疏矩阵)
        :param skip_check_transform_value: 是否跳过数值相等检测（目前只有FixInput模块要跳，其他都不会跳过）
        :param leak_check_transform_type: 弱化类型检测，比如将int32和int64都视为int类型
        :param leak_check_transform_value: 弱化检测值，将None视为相等
        :param copy_data: transform阶段是否要copy一次数据
        :param prefix: 是否为输出添加一个前缀,默认None,不添加
        :param kwargs:
        """
        super().__init__(name=name, **kwargs)
        self.input_col_names = None
        self.output_col_names = None
        self.transform_check_max_number_error = transform_check_max_number_error
        self.skip_check_transform_type = skip_check_transform_type
        self.skip_check_transform_value = skip_check_transform_value
        self.leak_check_transform_type = leak_check_transform_type
        self.leak_check_transform_value = leak_check_transform_value
        self.copy_data = copy_data
        self.prefix = prefix
        self.branch_pipes = []
        self.master_pipe = None  # 当前pipe作为branch pipe时所绑定的主master的pipe引用，这个对象无需保存，待申明时指定
        self.parent_pipe = None  # 父类对象不保存，实际部署时会重新申明使得保存的parent pipe失效

    def fit(self, s: dataframe_type, **kwargs):
        """
        fit:依次调用before_fit,_fit,after_fit

        :param s:
        :param kwargs:
        :return:
        """
        self.udf_fit(self.before_fit(s, **kwargs))
        return self.after_fit(s, **kwargs)

    def before_fit(self, s: dataframe_type, **kwargs) -> dataframe_type:
        """
        fit前预操作

        :param s:
        :param kwargs:
        :return:
        """
        s = super().before_fit(s, **kwargs)
        assert type(s) == dataframe_type
        self.input_col_names = s.columns.tolist()
        return s

    def udf_fit(self, s: dataframe_type, **kwargs):
        """

        :param s:
        :param kwargs:
        :return:
        """
        return self

    def after_fit(self, s: dataframe_type = None, **kwargs):
        """
        fit后操作

        :param s:
        :param kwargs:
        :return:
        """
        super().after_fit(s, **kwargs)
        # 执行branch pipe
        for branch_pipe in self.branch_pipes:
            branch_pipe.fit(copy.deepcopy(s), **kwargs)
        # 检验参数冲突
        self.get_params()
        return self

    def transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        """
        批量接口:依次调用before_transform,_transform,after_transform

        :param s:
        :param kwargs:
        :return:
        """
        return self.after_transform(self.udf_transform(self.before_transform(s, **kwargs), **kwargs), **kwargs)

    def before_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        """
        udf_transform之前调用

        :param s:
        :param kwargs:
        :return:
        """
        s = super().before_transform(s, **kwargs)
        assert type(s) == dataframe_type
        # 是否copy数据, copy可以避免修改input
        if self.copy_data:
            s_ = copy.deepcopy(s)
        else:
            s_ = s
        # 对其训练时的input columns
        input_cols = s.columns.tolist()
        if self.check_list_same(input_cols, self.input_col_names):
            return s_
        else:
            return s_[self.input_col_names]

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        """

        :param s:
        :param kwargs:
        :return:
        """
        return s

    def after_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        """
        udf_transform之后调用

        :param s:
        :param kwargs:
        :return:
        """
        s = super().after_transform(s, **kwargs)
        # 是否改名
        if self.prefix is not None:
            s.columns = ["{}_{}".format(self.prefix, col) for col in s.columns]
        # 保留output columns
        self.output_col_names = list(s.columns)
        # 执行branch pipe
        for branch_pipe in self.branch_pipes:
            branch_pipe.transform(copy.deepcopy(s), **kwargs)
        return s

    def transform_single(self, s: dict_type, **kwargs) -> dict_type:
        """
        单条数据接口:调用顺序before_transform_single,_transform_single,after_transform_single

        :param s:
        :param kwargs:
        :return:
        """
        return self.after_transform_single(
            self.udf_transform_single(self.before_transform_single(s, **kwargs), **kwargs), **kwargs)

    def before_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        """
        udf_transform_single之前调用

        :param s:
        :param kwargs:
        :return:
        """
        s = super().before_transform_single(s, **kwargs)
        assert type(s) == dict_type
        return self.extract_dict(s, self.input_col_names)

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        """
        单条数据接口，生产部署用

        :param s:
        :param kwargs:
        :return:
        """
        return s

    def after_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        """
        udf_transform_single之后调用

        :param s:
        :param kwargs:
        :return:
        """
        s = super().after_transform_single(s, **kwargs)
        # 改名
        if self.prefix is not None:
            new_s = dict()
            for col, value in s.items():
                new_s["{}_{}".format(self.prefix, col)] = value
            s = new_s
        # 输出指定columns
        s = self.extract_dict(s, self.output_col_names)
        # 执行branch pipe
        for branch_pipe in self.branch_pipes:
            branch_pipe.transform_single(copy.deepcopy(s), **kwargs)
        return s

    def get_params(self) -> dict_type:
        """

        :return:
        """
        # 逐步获取父类的_get_params
        current_class = self.__class__
        all_params = [self.udf_get_params()]
        while issubclass(current_class, TablePipeObjectBase):
            all_params.append(super(current_class, self).udf_get_params())
            self.check_key_conflict(all_params[-1], all_params[-2], current_class, current_class.__base__)
            current_class = current_class.__base__
        all_params.reverse()
        # 逆向聚合参数
        combine_params = dict()
        for params in all_params:
            combine_params.update(params)
        # 获取当前参数
        return combine_params

    def udf_get_params(self) -> dict_type:
        """

        :return:
        """
        return {"input_col_names": self.input_col_names,
                "output_col_names": self.output_col_names,
                "transform_check_max_number_error": self.transform_check_max_number_error,
                "skip_check_transform_type": self.skip_check_transform_type,
                "skip_check_transform_value": self.skip_check_transform_value,
                "copy_data": self.copy_data,
                "prefix": self.prefix,
                "branch_pipes": [branch_pipe.get_params() for branch_pipe in self.branch_pipes]}

    def set_params(self, params: dict_type):
        """

        :param params:
        :return:
        """
        # 逐步获取父类class
        current_class = self.__class__
        super_classes = []
        while issubclass(current_class, TablePipeObjectBase):
            super_classes.append(super(current_class, self))
            current_class = current_class.__base__
        super_classes.reverse()
        for super_class in super_classes:
            super_class.udf_set_params(params)
        # 设置当前类
        self.udf_set_params(params)

    def udf_set_params(self, params: dict_type):
        """
        :param params:
        :return:
        """
        self.input_col_names = params["input_col_names"]
        self.output_col_names = params["output_col_names"]
        self.transform_check_max_number_error = params["transform_check_max_number_error"]
        self.skip_check_transform_type = params["skip_check_transform_type"]
        self.skip_check_transform_value = params["skip_check_transform_value"]
        self.copy_data = params["copy_data"]
        self.prefix = params["prefix"]
        for index in range(len(params["branch_pipes"])):
            self.branch_pipes[index].set_params(params["branch_pipes"][index])

    @staticmethod
    def extract_dict(s: dict_type, keys: list) -> dict_type:
        """
        从s中抽取指定的keys

        :param s:
        :param keys:
        :return:
        """
        new_s = dict()
        for key in keys:
            new_s[key] = s[key]
        return new_s

    @staticmethod
    def check_list_same(list1: list, list2: list):
        """
        比较俩列表元素是否完全相同

        :param list1:
        :param list2:
        :return:
        """
        flag = True
        if len(list1) != len(list2):
            flag = False
        for idx in range(len(list1)):
            item1 = list1[idx]
            item2 = list2[idx]
            if item1 != item2:
                flag = False
                break
        return flag

    @staticmethod
    def check_key_conflict(param1: dict_type, param2: dict_type, class1, class2):
        """
        检测param1和param2中的key是否有冲突

        :param param1:
        :param param2:
        :param class1:
        :param class2:
        :return:
        """
        same_param_names = list(set(param1.keys()) & set(param2.keys()))
        if len(same_param_names) > 0:
            print("the {} and {} use same parameter names \033[1;43m[{}]\033[0m,please check if conflict,".format
                  (class1, class2, same_param_names))

    @staticmethod
    def get_col_type(pandas_col_type):
        """
        获取当前col的数据类型

        :param pandas_col_type:
        :return:
        """
        pandas_col_type = str(pandas_col_type).lower()
        if "int" in pandas_col_type:
            if "int8" in pandas_col_type:
                col_type = np.int8
            elif "int16" in pandas_col_type:
                col_type = np.int16
            elif "int32" in pandas_col_type:
                col_type = np.int32
            else:
                col_type = np.int64
        elif "float" in pandas_col_type:
            if "float16" in pandas_col_type:
                col_type = np.float16
            elif "float32" in pandas_col_type:
                col_type = np.float32
            else:
                col_type = np.float64
        else:
            col_type = str
        return col_type

    @staticmethod
    def get_match_values_index(x: series_type, values):
        """
        获取当前x的值是否匹配values
        """
        x_ = copy.deepcopy(x)
        index_map = dict()
        for value in values:
            index_map[value] = True
        index = x_.map(index_map).fillna(False)
        return index.astype(np.bool)

    def set_master_pipe(self, master_pipe):
        """
        设置master_pipe，这个函数是被pipeline调用的，当前pipe的角色是branch pipe

        :param master_pipe:
        :return:
        """
        self.master_pipe = master_pipe

    def get_master_pipe(self):
        """
        获取绑定的master pipe

        :return:
        """
        return self.master_pipe

    def set_branch_pipe(self, pipe_obj):
        """
        设置 branch pipe

        :param pipe_obj:
        :return:
        """
        pipe_obj.set_master_pipe(self)
        self.branch_pipes.append(pipe_obj)

    def get_branch_pipe(self, index):
        """
        获取指定index的branch pipe

        :param index:
        :return:
        """
        if type(index) == int:
            return self.branch_pipes[index]
        elif type(index) == str:
            for branch_pipe in self.branch_pipes:
                if branch_pipe.name == index:
                    return branch_pipe
        else:
            raise Exception(f"can't support the index:{index}")

    def remove_branch_pipe(self, index):
        """
        移除指定index的branch pipe

        :param index:
        :return:
        """
        hint_index = None
        if type(index) == int:
            hint_index = index
        elif type(index) == str:
            for branch_pipe in self.branch_pipes:
                if branch_pipe.name == index:
                    hint_index = index
                    break
        if hint_index is None:
            raise Exception(f"can't find the index:{index}")
        else:
            self.branch_pipes.pop(hint_index)

    def set_parent_pipe(self, parent_pipe=None):
        """
        设置前置的pipe，被pipeline调用

        :param parent_pipe:
        :return:
        """
        self.parent_pipe = parent_pipe

    def get_parent_pipe(self):
        """
        获取直接前置的pipe

        :return:
        """
        return self.parent_pipe

    def get_all_parent_pipes(self):
        """
        获取所有前置pipe 列表

        :return:
        """
        # 找到所有的父类(逆序保存)
        all_parent_pipes = []
        if self.parent_pipe is not None:
            all_parent_pipes.append(self.parent_pipe)
        while all_parent_pipes[-1].get_parent_pipe() is not None:
            all_parent_pipes.append(all_parent_pipes[-1].get_parent_pipe())
        # 改顺序
        all_parent_pipes.reverse()
        return all_parent_pipes

    def transform_all_parent(self, x, show_process=False, **kwargs):
        """
        将所有前置的pipe顺序执行一次transform

        :param x:
        :param show_process:
        :param kwargs:
        :return:
        """
        x_ = copy.deepcopy(x)
        all_parent_pipes = self.get_all_parent_pipes()
        for current_parent_pipe in tqdm(all_parent_pipes) if show_process else all_parent_pipes:
            if show_process:
                print(current_parent_pipe.name)
            x_ = current_parent_pipe.transform(x_, **kwargs)
        return x_

    def transform_single_all_parent(self, x, show_process=False, **kwargs):
        """
        将所有前置的pipe顺序执行一次transform_single

        :param x:
        :param show_process:
        :param kwargs:
        :return:
        """
        x_ = copy.deepcopy(x)
        all_parent_pipes = self.get_all_parent_pipes()
        for current_parent_pipe in tqdm(all_parent_pipes) if show_process else all_parent_pipes:
            if show_process:
                print(current_parent_pipe.name)
            x_ = current_parent_pipe.transform(x_, **kwargs)
        return x_



"""
自定义包
"""
from easymlops.table.core import *
import numpy as np
import scipy.stats as ss


class Normalization(TablePipeObjectBase):
    """
    归一化操作
    """

    def __init__(self, cols=None, normal_range=100, normal_type="cdf", std_range=10, **kwargs):
        """
        :param normal_range:
        :param normal_type: cdf,range
        :param std_range:
        """
        super().__init__(**kwargs)
        self.normal_range = normal_range
        self.normal_type = normal_type
        self.std_range = std_range
        self.mean_std = dict()
        self.cols = cols

    def udf_fit(self, s: dataframe_type, **kwargs):
        if str(self.cols).lower() in ["none", "all"]:
            self.cols = s.columns.tolist()
        if self.normal_type == "cdf":
            for col in self.cols:
                col_value = s[col]
                mean = np.median(col_value)
                std = np.std(col_value) * self.std_range
                self.mean_std[col] = (mean, std)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        if self.normal_type == "cdf":
            for col in self.cols:
                if col in self.mean_std:
                    s[col] = np.round(
                        ss.norm.cdf((s[col] - self.mean_std[col][0]) / self.mean_std[col][1]) * self.normal_range, 2)
        elif self.normal_type == "range":
            for col in self.cols:
                if col in self.mean_std:
                    s[col] = self.normal_range * s[col]
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        if self.normal_type == "cdf":
            for col in self.cols:
                if col in self.mean_std:
                    s[col] = np.round(
                        ss.norm.cdf((s[col] - self.mean_std[col][0]) / self.mean_std[col][1]) * self.normal_range, 2)
        elif self.normal_type == "range":
            for col in self.cols:
                if col in self.mean_std:
                    s[col] = self.normal_range * s[col]
        return s

    def udf_get_params(self) -> dict:
        return {"mean_std": self.mean_std, "normal_range": self.normal_range, "normal_type": self.normal_type,
                "cols": self.cols}

    def udf_set_params(self, params: dict):
        self.mean_std = params["mean_std"]
        self.normal_range = params["normal_range"]
        self.normal_type = params["normal_type"]
        self.cols = params["cols"]


class MapValues(TablePipeObjectBase):
    def __init__(self, map_values: dict = None, copy_data=True, **kwargs):
        super().__init__(copy_data=copy_data, **kwargs)
        self.map_values = map_values if map_values is not None else dict()

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in s.columns:
            if col in self.map_values:
                s[col] = np.round(s[col] / self.map_values[col][0] * self.map_values[col][1], 2)
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col in s.keys():
            if col in self.map_values:
                s[col] = np.round(s[col] / self.map_values[col][0] * self.map_values[col][1], 2)
        return s

    def udf_get_params(self) -> dict_type:
        return {"map_values": self.map_values}

    def udf_set_params(self, params: dict_type):
        self.map_values = params["map_values"]



from easymlops.table.core import *
from easymlops.table.utils import PandasUtils


class Parallel(TablePipeObjectBase):
    """
    并行模块:接受相同的数据，每个pipe object运行后合并(按col名覆盖)输出
    """

    def __init__(self, pipe_objects=None, drop_input_data=True, skip_check_transform_type=True, **kwargs):
        """

        :param pipe_objects:
        :param drop_input_data:
        :param skip_check_transform_type:
        :param kwargs:
        """
        super().__init__(skip_check_transform_type=skip_check_transform_type, **kwargs)
        self.pipe_objects = pipe_objects
        for obj in self.pipe_objects:
            # 设置parent
            obj.set_parent_pipe(self.get_parent_pipe())
        self.drop_input_data = drop_input_data

    def __getitem__(self, target_pipe_model):
        return self._match_index_pipe(match_index=target_pipe_model)

    def _match_index_pipe(self, match_index) -> TablePipeObjectBase:
        if type(match_index) == int:
            return self.pipe_objects[match_index]
        if type(match_index) == tuple:
            match_index = list(match_index)
        if type(match_index) == list:
            # 逐层获取
            pipes_ = []
            pipes_.extend(self.pipe_objects)
            current_pipe = None
            while len(match_index) > 0:
                index = match_index.pop(0)
                current_pipe = pipes_[index]
                pipes_ = []
                if issubclass(current_pipe.__class__, TablePipeLine):
                    pipes_.extend(current_pipe.models)
                if issubclass(current_pipe.__class__, easymlops.table.ensemble.Parallel):
                    pipes_.extend(current_pipe.pipe_objects)
            return current_pipe
        if type(match_index) == str:
            pipes_ = []
            pipes_.extend(self.pipe_objects)
            while len(pipes_) > 0:
                ipipe = pipes_.pop(0)
                if ipipe.name == match_index:
                    return ipipe
                if issubclass(ipipe.__class__, easymlops.table.ensemble.Parallel):
                    pipes_.extend(ipipe.pipe_objects)
                if issubclass(ipipe.__class__, TablePipeLine):
                    pipes_.extend(ipipe.models)
        raise Exception(f"can't match the index:{match_index}")

    def udf_fit(self, s: dataframe_type, **kwargs):
        for obj in self.pipe_objects:
            s_ = copy.deepcopy(s)  # 强制copy
            obj.fit(s_, **kwargs)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        dfs = []
        for index in range(len(self.pipe_objects)):
            s_ = copy.deepcopy(s)
            df = self.pipe_objects[index].transform(s_, **kwargs)
            df.index = s.index
            dfs.append(df)
        return PandasUtils.concat_duplicate_columns(dfs)

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        result = dict()
        for index in range(len(self.pipe_objects)):
            s_ = copy.deepcopy(s)
            result.update(self.pipe_objects[index].transform_single(s_, **kwargs))
        return result

    def transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        output = self.after_transform(self.udf_transform(self.before_transform(s, **kwargs), **kwargs), **kwargs)
        if self.drop_input_data:
            return output
        else:
            return PandasUtils.concat_duplicate_columns([s, output])

    def transform_single(self, s: dict_type, **kwargs) -> dict_type:
        output = self.after_transform_single(
            self.udf_transform_single(self.before_transform_single(s, **kwargs), **kwargs), **kwargs)
        if self.drop_input_data:
            return output
        else:
            s.update(output)
            return s

    def udf_get_params(self):
        return {"pipe_objects_params": [obj.get_params() for obj in self.pipe_objects],
                "pipe_objects": self.pipe_objects, "drop_input_data": self.drop_input_data}

    def udf_set_params(self, params: dict_type):
        self.drop_input_data = params["drop_input_data"]
        self.pipe_objects = params["pipe_objects"]
        pipe_objects_params = params["pipe_objects_params"]
        for i in range(len(pipe_objects_params)):
            pipe_params = pipe_objects_params[i]
            obj = self.pipe_objects[i]
            obj.set_params(pipe_params)



import scipy.sparse as sp
import numpy as np
import pandas as pd

dataframe_type = pd.DataFrame


class PandasUtils(object):
    """
    pandas操作工具箱
    """

    @staticmethod
    def pd2csr(data: dataframe_type):
        try:
            if hasattr(data, "sparse"):
                sparse_matrix = sp.csr_matrix(data.sparse.to_coo().tocsr(), dtype=np.float32)
            else:
                sparse_matrix = sp.csr_matrix(data, dtype=np.float32)
        except Exception as e:
            print("can't trans to sparse,because {}\n".format(e))
            sparse_matrix = data
        return sparse_matrix

    @staticmethod
    def pd2dense(data: dataframe_type):
        return pd.DataFrame(data.values, index=data.index)

    @staticmethod
    def concat_duplicate_columns(dfs, keep="last"):
        """
        keep:last/相同column保留最后一个，first/相同column保留第一个
        """

        def flatten(nest_list: list):
            return [j for i in nest_list for j in flatten(i)] if isinstance(nest_list, list) else [nest_list]

        # 所有列保证相同的index
        df_start = dfs[0]
        total_columns = [df_start.columns.tolist()]
        for df in dfs[1:]:
            df.index = df_start.index
            total_columns.append(df.columns.tolist())
        # 删除重复列
        if keep == "first":
            dfs.reverse()
            total_columns.reverse()
        # 找出冲突列
        hint_columns = []
        for index in range(len(total_columns) - 1):
            current_columns = total_columns[index]
            remind_columns = flatten(total_columns[index + 1:])
            hint_columns.append(list(set(current_columns) & set(remind_columns)))
        hint_columns.append([])
        # 删除
        for index in range(len(dfs)):
            df = dfs[index]
            hint_column = hint_columns[index]
            for col in hint_column:
                del df[col]
        if keep == "first":
            dfs.reverse()
        return pd.concat(dfs, axis=1)



from easymlops.table.core import *
from ..encoding import WOEEncoding
from ..preprocessing import FillNa


class FilterBase(TablePipeObjectBase):
    """
    过滤式特征选择的基础类
    """

    def __init__(self, cols="all", as_null_values=None, min_threshold=None, max_threshold=None,
                 skip_check_transform_type=True, skip_check_transform_value=True,
                 **kwargs):
        """

        :param cols: 用于特征选择的cols
        :param as_null_values: 等视为Null的值
        :param min_threshold: 最小阈值,None表示跳过
        :param max_threshold: 最大阈值,None表示跳过
        :param skip_check_transform_type:跳过类型检测
        :param skip_check_transform_value:跳过数值检测
        :param kwargs:
        """
        super().__init__(skip_check_transform_value=skip_check_transform_value,
                         skip_check_transform_type=skip_check_transform_type, **kwargs)
        self.cols = cols
        self.min_threshold = min_threshold
        self.max_threshold = max_threshold
        if self.min_threshold is None:
            self.min_threshold = np.finfo(np.float64).min
        if self.max_threshold is None:
            self.max_threshold = np.finfo(np.float64).max
        self.as_null_values = as_null_values
        if self.as_null_values is None:
            self.as_null_values = GlobalNullValues
        # filter_value_dist分布
        self.filter_value_dist = dict()
        self.selected_cols = None

    def get_filter_value_dist(self):
        return self.filter_value_dist

    def show_detail(self):
        return pd.DataFrame([self.filter_value_dist])

    def before_fit(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s = super().before_fit(s, **kwargs)
        if str(self.cols).lower() in ["none", "all", "null", "nan"]:
            self.cols = self.input_col_names
        assert type(self.cols) == list and type(self.cols[0]) == str
        return s

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        return s[self.selected_cols]

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        return self.extract_dict(s, self.selected_cols)

    def udf_get_params(self):
        return {"cols": self.cols, "min_threshold": self.min_threshold,
                "max_threshold": self.max_threshold, "as_null_values": self.as_null_values,
                "filter_value_dist": self.filter_value_dist, "selected_cols": self.selected_cols}

    def udf_set_params(self, params: dict_type):
        self.cols = params["cols"]
        self.min_threshold = params["min_threshold"]
        self.max_threshold = params["max_threshold"]
        self.as_null_values = params["as_null_values"]
        self.filter_value_dist = params["filter_value_dist"]
        self.selected_cols = params["selected_cols"]


class MissRateFilter(FilterBase):
    """
    缺失率过滤
    """

    def __init__(self, max_threshold=0.95, min_threshold=None, **kwargs):
        super().__init__(max_threshold=max_threshold, min_threshold=min_threshold, **kwargs)

    def udf_fit(self, s: dataframe_type, **kwargs):
        self.filter_value_dist = dict()
        for col in self.cols:
            hint_null_index = self.get_match_values_index(s[col], self.as_null_values)
            miss_rate = np.sum(hint_null_index) / len(s)
            self.filter_value_dist[col] = miss_rate
        self.selected_cols = []
        for col in s.columns:
            miss_rate = self.filter_value_dist.get(col)
            if miss_rate is None or col not in self.cols:
                self.selected_cols.append(col)
            else:
                if self.min_threshold <= miss_rate <= self.max_threshold:
                    self.selected_cols.append(col)
        return self

    def udf_get_params(self):
        return {}


class VarianceFilter(FilterBase):
    """
    方差值过滤
    """

    def __init__(self, max_threshold=None, min_threshold=1e-3, **kwargs):
        super().__init__(max_threshold=max_threshold, min_threshold=min_threshold, **kwargs)

    def udf_fit(self, s: dataframe_type, **kwargs):
        self.filter_value_dist = s[self.cols].var().to_dict()
        self.selected_cols = []
        for col in s.columns:
            miss_rate = self.filter_value_dist.get(col)
            if miss_rate is None or col not in self.cols:
                self.selected_cols.append(col)
            else:
                if self.min_threshold <= miss_rate <= self.max_threshold:
                    self.selected_cols.append(col)
        return self

    def udf_get_params(self):
        return {}


class PersonCorrFilter(FilterBase):
    """
    person相关性绝对值过滤
    """

    def __init__(self, y=None, max_threshold=None, min_threshold=0.01, **kwargs):
        super().__init__(max_threshold=max_threshold, min_threshold=min_threshold, **kwargs)
        self.y = y

    def udf_fit(self, s: dataframe_type, **kwargs):
        assert self.y is not None and len(self.y) == len(s)
        self.filter_value_dist = s[self.cols].corrwith(self.y).to_dict()
        self.selected_cols = []
        for col in s.columns:
            miss_rate = self.filter_value_dist.get(col)
            if miss_rate is None or col not in self.cols:
                self.selected_cols.append(col)
            else:
                if self.min_threshold <= abs(miss_rate) <= self.max_threshold:
                    self.selected_cols.append(col)
        return self

    def udf_get_params(self):
        return {}


class Chi2Filter(FilterBase):
    """
    卡方过滤
    """

    def __init__(self, y=None, max_threshold=None, min_threshold=0.01, **kwargs):
        super().__init__(max_threshold=max_threshold, min_threshold=min_threshold, **kwargs)
        self.y = y

    def udf_fit(self, s: dataframe_type, **kwargs):
        from sklearn.feature_selection import chi2
        assert self.y is not None and len(self.y) == len(s)
        self.filter_value_dist = dict()
        for idx, filter_value in enumerate(chi2(s[self.cols], self.y)[0]):
            self.filter_value_dist[self.cols[idx]] = filter_value
        del self.y
        self.selected_cols = []
        for col in s.columns:
            miss_rate = self.filter_value_dist.get(col)
            if miss_rate is None or col not in self.cols:
                self.selected_cols.append(col)
            else:
                if self.min_threshold <= abs(miss_rate) <= self.max_threshold:
                    self.selected_cols.append(col)
        return self

    def udf_get_params(self):
        return {}


class PValueFilter(FilterBase):
    """
    P值过滤
    """

    def __init__(self, y=None, max_threshold=None, min_threshold=None, **kwargs):
        super().__init__(max_threshold=max_threshold, min_threshold=min_threshold, **kwargs)
        self.y = y

    def udf_fit(self, s: dataframe_type, **kwargs):
        from sklearn.feature_selection import chi2
        assert self.y is not None and len(self.y) == len(s)
        self.filter_value_dist = dict()
        for idx, filter_value in enumerate(chi2(s[self.cols], self.y)[1]):
            self.filter_value_dist[self.cols[idx]] = filter_value
        del self.y
        self.selected_cols = []
        for col in s.columns:
            miss_rate = self.filter_value_dist.get(col)
            if miss_rate is None or col not in self.cols:
                self.selected_cols.append(col)
            else:
                if self.min_threshold <= abs(miss_rate) <= self.max_threshold:
                    self.selected_cols.append(col)
        return self

    def udf_get_params(self):
        return {}


class MutualInfoFilter(FilterBase):
    """
    互信息过滤
    """

    def __init__(self, y=None, max_threshold=None, min_threshold=None, **kwargs):
        super().__init__(max_threshold=max_threshold, min_threshold=min_threshold, **kwargs)
        self.y = y

    def udf_fit(self, s: dataframe_type, **kwargs):
        from sklearn.feature_selection import mutual_info_classif
        assert self.y is not None and len(self.y) == len(s)
        self.filter_value_dist = dict()
        for idx, filter_value in enumerate(mutual_info_classif(s[self.cols], self.y)):
            self.filter_value_dist[self.cols[idx]] = filter_value
        del self.y
        self.selected_cols = []
        for col in s.columns:
            miss_rate = self.filter_value_dist.get(col)
            if miss_rate is None or col not in self.cols:
                self.selected_cols.append(col)
            else:
                if self.min_threshold <= abs(miss_rate) <= self.max_threshold:
                    self.selected_cols.append(col)
        return self

    def udf_get_params(self):
        return {}


class IVFilter(FilterBase):
    """
    IV过滤
    """

    def __init__(self, y=None, max_threshold=None, min_threshold=0.02, **kwargs):
        super().__init__(max_threshold=max_threshold, min_threshold=min_threshold, **kwargs)
        self.y = y

    def udf_fit(self, s: dataframe_type, **kwargs):
        assert self.y is not None and len(self.y) == len(s)
        self.filter_value_dist = \
            WOEEncoding(y=self.y, cols=self.cols).fit(s, **kwargs).show_detail()[["col", "iv"]].groupby("col").agg(
                {"iv": "sum"}).to_dict()["iv"]
        self.selected_cols = []
        for col in s.columns:
            miss_rate = self.filter_value_dist.get(col)
            if miss_rate is None or col not in self.cols:
                self.selected_cols.append(col)
            else:
                if self.min_threshold <= abs(miss_rate) <= self.max_threshold:
                    self.selected_cols.append(col)
        return self

    def udf_get_params(self):
        return {}


class PSIFilter(FilterBase):
    """
    psi过滤
    """

    def __init__(self, oot_x=None, max_threshold=1, min_threshold=None, save_detail=True, fill_na=True, **kwargs):
        super().__init__(max_threshold=max_threshold, min_threshold=min_threshold, **kwargs)
        self.oot_x = oot_x
        self.save_detail = save_detail
        self.dist_detail = []
        self.fill_na = fill_na

    def show_detail(self):
        return pd.DataFrame(data=self.dist_detail,
                            columns=["col", "bin_value", "ins_num", "ins_rate", "oot_num", "oot_rate", "psi"])

    def udf_fit(self, s: dataframe_type, **kwargs):
        oot_x = self.transform_all_parent(self.oot_x)
        if self.fill_na:
            fill_na_model = FillNa(cols=self.cols)
            s = fill_na_model.fit(s, **kwargs).transform(s, **kwargs)
            oot_x = fill_na_model.transform(oot_x, **kwargs)
        total_ins_num = len(s)
        total_oot_num = len(oot_x)
        self.dist_detail = []
        for col in self.cols:
            for bin_value in s[col].unique().tolist():
                ins_num = np.sum(s[col] == bin_value)
                oot_num = np.sum(oot_x[col] == bin_value)
                ins_rate = ins_num / total_ins_num
                oot_rate = oot_num / total_oot_num
                if ins_num == 0 or oot_num == 0:
                    psi = 0
                else:
                    psi = (ins_rate - oot_rate) * np.log(ins_rate / oot_rate)
                self.dist_detail.append([col, bin_value, ins_num, ins_rate, oot_num, oot_rate, psi])
        self.filter_value_dist = self.show_detail()[["col", "psi"]].groupby("col").agg(
            {"psi": "sum"}).to_dict()["psi"]
        self.selected_cols = []
        for col in s.columns:
            miss_rate = self.filter_value_dist.get(col)
            if miss_rate is None or col not in self.cols:
                self.selected_cols.append(col)
            else:
                if self.min_threshold <= abs(miss_rate) <= self.max_threshold:
                    self.selected_cols.append(col)

        del self.oot_x
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        return s[self.selected_cols]

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        return self.extract_dict(s, self.selected_cols)

    def udf_get_params(self):
        params = {"save_detail": self.save_detail}
        if self.save_detail:
            params["dist_detail"] = self.dist_detail
        return params

    def udf_set_params(self, params: dict_type):
        self.save_detail = params["save_detail"]
        if self.save_detail:
            self.dist_detail = params["dist_detail"]
        else:
            self.dist_detail = []



from .core import *


class Replace(PreprocessBase):
    """
    对单个col进行的操作
    """

    def __init__(self, cols="all", source_values=None, target_value="", **kwargs):
        """

        :param cols:
        :param source_values:
        :param target_value:
        :param kwargs:
        """
        super().__init__(cols=cols, **kwargs)
        self.source_values = source_values
        if source_values is None:
            self.source_values = []
        assert type(self.source_values) == list
        self.target_value = target_value

    def apply_function_single(self, col: str, x):
        if x in self.source_values:
            return self.target_value
        else:
            return x

    def apply_function_series(self, col: str, x: series_type):
        index = self.get_match_values_index(x, self.source_values)
        x.loc[index] = self.target_value
        return x

    def udf_get_params(self):
        return {"source_values": self.source_values, "target_value": self.target_value}

    def udf_set_params(self, params: dict_type):
        self.source_values = params["source_values"]
        self.target_value = params["target_value"]


class ClipString(PreprocessBase):
    """
    截图字符串的指定位置
    """

    def __init__(self, cols="all", default_clip_index: tuple = (0, None), clip_detail: dict = None,
                 **kwargs):
        super().__init__(cols, **kwargs)
        self.default_clip_index = default_clip_index

        self.clip_detail = clip_detail

    def udf_fit(self, s: dataframe_type, **kwargs):
        if self.clip_detail is None:
            self.clip_detail = dict()
            for col, _ in self.cols:
                self.clip_detail[col] = self.default_clip_index
        else:
            new_cols = []
            for col, new_col in self.cols:
                if col in self.clip_detail.keys():
                    new_cols.append((col, new_col))
            self.cols = new_cols
        return self

    def apply_function_series(self, col: str, x: series_type):
        clip_index = slice(*self.clip_detail[col])
        try:
            return x.astype(str).str[clip_index]
        except Exception as e:
            print("module:{},in function:{},raise the exception:{}".format(self.name, "apply_function_series", e))
            return x.astype(str).apply(lambda x_: self.apply_function_single(col, x_))

    def apply_function_single(self, col: str, x):
        clip_index = slice(*self.clip_detail[col])
        try:
            return str(x)[clip_index]
        except Exception as e:
            print("module:{},in function:{},raise the exception:{}".format(self.name, "apply_function_single", e))
            return ""

    def udf_get_params(self):
        return {"clip_detail": self.clip_detail, "default_clip_index": self.default_clip_index}

    def udf_set_params(self, params: dict):
        self.default_clip_index = params["default_clip_index"]
        self.clip_detail = params["clip_detail"]


class FillNa(PreprocessBase):
    """
    对cols中空进行填充，有三个优先级 \n
    1.fill_detail不为空,则处理fill_detail所指定的填充值，格式为fill_detail={"col1":1,"col2":"miss"}表示将col1中的空填充为1，col2中的填充为"miss" \n
    2.fill_detail为空时，接着看fill_mode，可选项有mean:表示均值填充，median:中位数填充，mode:众数填充(对number和category类型均可) \n
    3.当fill_detail和fill_mode均为空时，将number数据类型用fill_number_value填充，category数据类型用fill_category_value填充
    """

    def __init__(self, cols="all", fill_mode=None, fill_number_value=0, fill_category_value="nan", fill_detail=None,
                 default_null_values=None,
                 **kwargs):
        super().__init__(cols=cols, **kwargs)
        if default_null_values is None:
            default_null_values = GlobalNullValues
        self.fill_number_value = fill_number_value
        self.fill_category_value = fill_category_value
        self.fill_mode = fill_mode
        self.fill_detail = fill_detail
        self.default_null_values = default_null_values

    def udf_fit(self, s: dataframe_type, **kwargs):
        if self.fill_detail is None:
            self.fill_detail = dict()
            if self.fill_mode is not None:
                for col, _ in self.cols:
                    if self.fill_mode == "mean":
                        self.fill_detail[col] = s[col].mean()
                    elif self.fill_mode == "median":
                        self.fill_detail[col] = s[col].median()
                    elif self.fill_mode == "mode":
                        self.fill_detail[col] = s[col].mode()[0]
                    else:
                        raise Exception("fill_model should be [mean,median,mode]")
            else:
                for col, _ in self.cols:
                    if "int" in str(s[col].dtype).lower() or "float" in str(s[col].dtype).lower():
                        self.fill_detail[col] = self.fill_number_value
                    else:
                        self.fill_detail[col] = self.fill_category_value
        else:
            new_cols = []
            for col, new_col in self.cols:
                if col in self.fill_detail.keys():
                    new_cols.append((col, new_col))
            self.cols = new_cols
        return self

    def apply_function_single(self, col, x):
        fill_value = self.fill_detail.get(col)
        if str(x).lower() in self.default_null_values:
            x = fill_value
        return x

    def apply_function_series(self, col: str, x: series_type):
        fill_value = self.fill_detail.get(col)
        index = self.get_match_values_index(x, self.default_null_values)
        x.loc[index] = fill_value
        return x

    def udf_get_params(self) -> dict_type:
        return {"fill_detail": self.fill_detail, "default_null_values": self.default_null_values}

    def udf_set_params(self, params: dict):
        self.fill_detail = params["fill_detail"]
        self.default_null_values = params["default_null_values"]


class IsNull(PreprocessBase):
    """
    判断是否为空
    """

    def __init__(self, cols="all", default_null_values=None,
                 **kwargs):
        super().__init__(cols=cols, **kwargs)
        if default_null_values is None:
            default_null_values = GlobalNullValues
        self.default_null_values = default_null_values
        self.null_value_map = dict()
        for null_value in self.default_null_values:
            self.null_value_map[null_value] = 1

    def udf_fit(self, s: dataframe_type, **kwargs):
        new_cols = []
        for col, new_col in self.cols:
            if col == new_col:
                new_cols.append((col, col + "_is_null"))
        self.cols = new_cols
        return self

    def apply_function_single(self, col, x):
        if str(x).lower() in self.default_null_values:
            return 1
        else:
            return 0

    def apply_function_series(self, col: str, x: series_type):
        return x.map(self.null_value_map).fillna(0).astype(np.uint8)

    def udf_get_params(self) -> dict_type:
        return {"default_null_values": self.default_null_values, "null_value_map": self.null_value_map}

    def udf_set_params(self, params: dict):
        self.default_null_values = params["default_null_values"]
        self.null_value_map = params["null_value_map"]


class IsNotNull(PreprocessBase):
    """
    判断是否不为空
    """

    def __init__(self, cols="all", default_null_values=None,
                 **kwargs):
        super().__init__(cols=cols, **kwargs)
        if default_null_values is None:
            default_null_values = GlobalNullValues
        self.default_null_values = default_null_values
        self.null_value_map = dict()
        for null_value in self.default_null_values:
            self.null_value_map[null_value] = 0

    def udf_fit(self, s: dataframe_type, **kwargs):
        new_cols = []
        for col, new_col in self.cols:
            if col == new_col:
                new_cols.append((col, col + "_is_not_null"))
        self.cols = new_cols
        return self

    def apply_function_single(self, col, x):
        if str(x).lower() in self.default_null_values:
            return 0
        else:
            return 1

    def apply_function_series(self, col: str, x: series_type):
        return x.map(self.null_value_map).fillna(1).astype(np.uint8)

    def udf_get_params(self) -> dict_type:
        return {"default_null_values": self.default_null_values, "null_value_map": self.null_value_map}

    def udf_set_params(self, params: dict):
        self.default_null_values = params["default_null_values"]
        self.null_value_map = params["null_value_map"]


class TransToCategory(PreprocessBase):
    """
    将cols中的数据转换为category字符类型
    """

    def __init__(self, cols="all", **kwargs):
        super().__init__(cols=cols, **kwargs)

    def apply_function_single(self, col: str, x):
        return str(x)

    def apply_function_series(self, col: str, x: series_type):
        return x.astype(str)

    def udf_get_params(self) -> dict_type:
        return {}


class TransToFloat(PreprocessBase):

    def __init__(self, cols="all", nan_fill_value=0, **kwargs):
        super().__init__(cols=cols, **kwargs)
        self.nan_fill_value = nan_fill_value

    def apply_function_single(self, col: str, x):
        x = pd.to_numeric(x, errors="coerce")
        if "nan" in str(x).lower():
            x = self.nan_fill_value
        return x

    def apply_function_series(self, col: str, x: series_type):
        return pd.to_numeric(x, errors="coerce").fillna(self.nan_fill_value)

    def udf_get_params(self) -> dict_type:
        return {"nan_fill_value": self.nan_fill_value}

    def udf_set_params(self, params: dict):
        self.nan_fill_value = params["nan_fill_value"]


class TransToInt(PreprocessBase):
    """
    将cols中的数据转换为int类型，对于处理异常的情况用nan_fill_value填充
    """

    def __init__(self, cols="all", nan_fill_value=0, **kwargs):
        super().__init__(cols=cols, **kwargs)
        self.nan_fill_value = nan_fill_value

    def apply_function_single(self, col: str, x):
        x = pd.to_numeric(x, errors="coerce")
        if "nan" in str(x).lower():
            x = self.nan_fill_value
        return np.int64(x)

    def apply_function_series(self, col: str, x: series_type):
        return pd.to_numeric(x, errors="coerce").fillna(self.nan_fill_value).astype(np.int64)

    def udf_get_params(self) -> dict_type:
        return {"nan_fill_value": self.nan_fill_value}

    def udf_set_params(self, params: dict):
        self.nan_fill_value = params["nan_fill_value"]


class TransToLower(PreprocessBase):
    """
    将字符中的所有英文字符转小写
    """

    def apply_function_single(self, col: str, x):
        return str(x).lower()

    def apply_function_series(self, col: str, x: series_type):
        return x.astype(str).str.lower()

    def udf_get_params(self):
        return {}


class TransToUpper(PreprocessBase):
    """
    将字符中的所有英文字符转大写
    """

    def apply_function_single(self, col: str, x):
        return str(x).upper()

    def apply_function_series(self, col: str, x: series_type):
        return x.astype(str).str.lower()

    def udf_get_params(self):
        return {}


class Abs(PreprocessBase):
    """
    取绝对值
    """

    def apply_function_single(self, col: str, x):
        return abs(x)

    def apply_function_series(self, col: str, x: series_type):
        return abs(x)

    def udf_get_params(self):
        return {}


class MapValues(PreprocessBase):
    """
    基本结构表示:[(["a","b",1],"c"), \n
               ("[4,5)",1), \n
               0] \n
    []中: \n
    1.第一个表示离散型映射，把"a","b",1映射为"c" \n
    2.第二个表示范围内映射，把4<=x<5的值，映射为1，其中[,]表示闭区间,(,)表示开区间 \n
    3.列表的最后一个是没有匹配上的设置的默认值 \n
    """

    def __init__(self, cols="all", map_detail=None, default_map=None, **kwargs):
        super().__init__(cols=cols, **kwargs)
        self.default_map = default_map
        self.map_detail = map_detail

    def udf_fit(self, s: dataframe_type, **kwargs):
        if self.map_detail is None:
            self.map_detail = dict()
            for col, _ in self.cols:
                self.map_detail[col] = self.default_map
        else:
            new_cols = []
            for col, new_col in self.cols:
                if col in self.map_detail.keys():
                    new_cols.append((col, new_col))
            self.cols = new_cols
        return self

    @staticmethod
    def get_match_range_index(x, range_str):
        x_ = pd.to_numeric(x, errors="coerce")
        min_value, max_value = range_str.split(',')
        if '[' in min_value:
            min_value = min_value.strip().replace('[', '')
            min_value = float(min_value.strip())
            match_index1 = min_value <= x_
        else:
            min_value = min_value.strip().replace('(', '')
            min_value = float(min_value.strip())
            match_index1 = min_value < x_

        if ']' in max_value:
            max_value = max_value.strip().replace(']', '')
            max_value = float(max_value.strip())
            match_index2 = x_ <= max_value
        else:
            max_value = max_value.strip().replace(')', '')
            max_value = float(max_value.strip())
            match_index2 = x_ < max_value
        return match_index1 & match_index2

    def apply_function_series(self, col: str, x: series_type):
        map_detail = self.map_detail[col]
        default_value = map_detail[-1]
        if type(default_value) != tuple:
            return_default_value = True
            map_detail_ = map_detail[:-1]
        else:
            return_default_value = False
            map_detail_ = map_detail
        match_index_with_target_value = []
        # 分别获取每种映射的match_index和target_value
        # 这里注意不要逐步替换，避免同一个值被转换2次
        for match_pattern, target_value in map_detail_:
            if type(match_pattern) == list:
                match_index = self.get_match_values_index(x, match_pattern)
            else:
                match_index = self.get_match_range_index(x, match_pattern)
            match_index_with_target_value.append((match_index, target_value))

        # 定一个全False的index
        total_match_index = match_index_with_target_value[0][0] != match_index_with_target_value[0][0]
        for match_index, target_value in match_index_with_target_value:
            x.loc[match_index] = target_value
            total_match_index = total_match_index | match_index
        if return_default_value:
            x.loc[~total_match_index] = default_value  # 都未被匹配的设置默认值
        return x

    def apply_function_single(self, col: str, x):
        map_detail = self.map_detail[col]
        default_value = map_detail[-1]
        if type(default_value) != tuple:
            return_default_value = True
            map_detail_ = map_detail[:-1]
        else:
            return_default_value = False
            map_detail_ = map_detail
        for match_pattern, target_value in map_detail_:
            if type(match_pattern) == list:
                if x in match_pattern:
                    return target_value
            elif self.get_match_range_index(x, match_pattern):
                return target_value
        if return_default_value:
            return default_value
        else:
            return x

    def udf_get_params(self) -> dict_type:
        return {"map_detail": self.map_detail}

    def udf_set_params(self, params: dict):
        self.map_detail = params["map_detail"]


class Clip(PreprocessBase):
    """
    对数值数据进行盖帽操作，有三个优先级 \n
    1.clip_detail不为空，clip_detail={"col1":(-1,1),"col2":(0,1)}表示将col1中<=-1的值设置为-1，>=1的值设置为1，将col2中<=0的值设置为0，>=1的值设置为1 \n
    2.clip_detail为空，percent_range不为空，percent_range=(1,99)表示对所有cols，对最小的1%和最高的99%数据进行clip \n
    3.clip_detail和percent_range均为空时，default_clip=(0,1)表示对所有cols，<=0的设置为0，>=1的设置为1 \n
    """

    def __init__(self, cols="all", default_clip=None, clip_detail=None, percent_range=None, **kwargs):
        """
       优先级clip_detail>percent_range>default_clip
        """
        super().__init__(cols=cols, **kwargs)
        self.default_clip = default_clip
        self.clip_detail = clip_detail
        self.percent_range = percent_range

    def apply_function_single(self, col: str, x):
        clip_detail_ = self.clip_detail.get(col, (None, None))
        x = np.clip(pd.to_numeric(x, errors="coerce"), a_min=clip_detail_[0], a_max=clip_detail_[1])
        return x

    def apply_function_series(self, col: str, x: series_type):
        clip_detail_ = self.clip_detail.get(col, (None, None))
        x = np.clip(pd.to_numeric(x, errors="coerce"), a_min=clip_detail_[0], a_max=clip_detail_[1])
        return x

    def udf_fit(self, s: dataframe_type, **kwargs):
        if self.clip_detail is None:
            self.clip_detail = dict()
            if self.percent_range is None:
                for col, _ in self.cols:
                    self.clip_detail[col] = self.default_clip
            else:
                for col, _ in self.cols:
                    if self.percent_range[0] is not None:
                        low = np.percentile(s[col], self.percent_range[0])
                    else:
                        low = None
                    if self.percent_range[1] is not None:
                        top = np.percentile(s[col], self.percent_range[1])
                    else:
                        top = None
                    self.clip_detail[col] = (low, top)
        else:
            new_cols = []
            for col, new_col in self.cols:
                if col in self.clip_detail.keys():
                    new_cols.append((col, new_col))
            self.cols = new_cols
        return self

    def udf_get_params(self) -> dict_type:
        return {"clip_detail": self.clip_detail}

    def udf_set_params(self, params: dict):
        self.clip_detail = params["clip_detail"]


class MinMaxScaler(PreprocessBase):
    """
    对cols进行最大最小归一化:(x-min)/(max-min) \n
    但如果max和min相等，则设置为1 \n
    """

    def __init__(self, decimals=2, **kwargs):
        super().__init__(**kwargs)
        self.min_max_detail = dict()
        self.col_types = dict()
        self.decimals = decimals

    def _check_min_max_equal(self, col, min_value, max_value):
        if min_value == max_value:
            print("({}), in  column \033[1;43m{}\033[0m ,min value and max value has the same value:{},"
                  "the finally result will be set 1".format(self.name, col, min_value))

    def show_detail(self):
        data = []
        for col, value in self.min_max_detail.items():
            min_value, max_value = value
            data.append([col, min_value, max_value])
        return pd.DataFrame(data=data, columns=["col", "min_value", "max_value"])

    def udf_fit(self, s: dataframe_type, **kwargs):
        for col, _ in self.cols:
            col_value = s[col].astype(np.float64)
            min_value = np.min(col_value)
            max_value = np.max(col_value)
            self.min_max_detail[col] = (min_value, max_value)
            self.col_types[col] = self.get_col_type(s[col])
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col, new_col in self.cols:
            min_value, max_value = self.min_max_detail.get(col)
            col_type = np.float64
            self._check_min_max_equal(col, min_value, max_value)
            if min_value == max_value:
                s[new_col] = 1
            else:
                s[new_col] = (s[col].astype(col_type) - col_type(min_value)) / (
                        col_type(max_value) - col_type(min_value))
                s[new_col] = np.round(s[new_col], self.decimals)
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col, new_col in self.cols:
            min_value, max_value = self.min_max_detail.get(col)
            col_type = np.float64
            self._check_min_max_equal(col, min_value, max_value)
            if min_value == max_value:
                s[new_col] = 1
            else:
                s[new_col] = (col_type(s[col]) - col_type(min_value)) / (col_type(max_value) - col_type(min_value))
                s[new_col] = np.round(s[new_col], self.decimals)
        return s

    def udf_get_params(self):
        return {"min_max_detail": self.min_max_detail, "col_types": self.col_types, "decimals": self.decimals}

    def udf_set_params(self, params: dict):
        self.min_max_detail = params["min_max_detail"]
        self.col_types = params["col_types"]
        self.decimals = params["decimals"]


class Normalizer(PreprocessBase):
    """
    对cols进行标准化:(x-mean)/std \n
    但如果std=0，则直接设置为1 \n
    """

    def __init__(self, decimals=2, **kwargs):
        super().__init__(**kwargs)
        self.mean_std_detail = dict()
        self.col_types = dict()
        self.decimals = decimals

    def _check_std(self, col, std):
        if std == 0:
            print("({}), in  column \033[1;43m{}\033[0m ,the std is 0,"
                  "the finally result will be set 1".format(self.name, col))

    def show_detail(self):
        data = []
        for col, value in self.mean_std_detail.items():
            mean_value, std_value = value
            data.append([col, mean_value, std_value])
        return pd.DataFrame(data=data, columns=["col", "mean_value", "std_value"])

    def udf_fit(self, s: dataframe_type, **kwargs):
        for col, _ in self.cols:
            col_value = s[col].astype(np.float64)
            mean_value = np.mean(col_value)
            std_value = np.std(col_value)
            self.mean_std_detail[col] = (mean_value, std_value)
            self.col_types[col] = self.get_col_type(s[col])
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col, new_col in self.cols:
            mean_value, std_value = self.mean_std_detail.get(col)
            col_type = np.float64
            self._check_std(col, std_value)
            if std_value == 0:
                s[new_col] = 1
            else:
                s[new_col] = (s[col].astype(col_type) - col_type(mean_value)) / col_type(std_value)
                s[new_col] = np.round(s[new_col], self.decimals)
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col, new_col in self.cols:
            mean_value, std_value = self.mean_std_detail.get(col)
            col_type = np.float64
            self._check_std(col, std_value)
            if std_value == 0:
                s[new_col] = 1
            else:
                s[new_col] = (col_type(s[col]) - col_type(mean_value)) / col_type(std_value)
                s[new_col] = np.round(s[new_col], self.decimals)
        return s

    def udf_get_params(self):
        return {"mean_std_detail": self.mean_std_detail, "col_types": self.col_types, "decimals": self.decimals}

    def udf_set_params(self, params: dict):
        self.mean_std_detail = params["mean_std_detail"]
        self.col_types = params["col_types"]
        self.decimals = params["decimals"]


class Bins(PreprocessBase):
    """
    对cols进行分箱 \n
    """

    def __init__(self, n_bins=10, strategy="quantile", **kwargs):
        """
        :param n_bins: 分箱数
        :param strategy: uniform/quantile/kmeans分别表示等距/等位/kmeans聚类分箱
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.n_bins = n_bins
        if self.n_bins < 2 ** 8:
            self.col_type = np.uint8
        elif self.n_bins < 2 ** 16:
            self.col_type = np.uint16
        elif self.n_bins < 2 ** 32:
            self.col_type = np.uint32
        else:
            self.col_type = np.uint64
        self.strategy = strategy
        self.bin_detail = dict()

    def udf_fit(self, s: dataframe_type, **kwargs):
        from sklearn.preprocessing import KBinsDiscretizer
        for col, _ in self.cols:
            bin_model = KBinsDiscretizer(n_bins=self.n_bins, encode="ordinal", strategy=self.strategy)
            bin_model.fit(s[[col]])
            self.bin_detail[col] = bin_model
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col, new_col in self.cols:
            bin_model = self.bin_detail[col]
            s[new_col] = bin_model.transform(s[[col]]).astype(self.col_type)
        return s

    def transform_single(self, s: dict_type, **kwargs) -> dict_type:
        input_dataframe = pd.DataFrame([s])
        return self.transform(input_dataframe, **kwargs).to_dict("record")[0]

    def udf_get_params(self):
        return {"n_bins": self.n_bins, "strategy": self.strategy, "bin_detail": self.bin_detail,
                "col_type": self.col_type}

    def udf_set_params(self, params: dict_type):
        self.bin_detail = params["bin_detail"]
        self.n_bins = params["n_bins"]
        self.strategy = params["strategy"]
        self.col_type = params["col_type"]



import easymlops.table.ensemble
from easymlops.table.core import TablePipeObjectBase, dict_type
import copy
from tqdm import tqdm
import pickle


class TablePipeLine(TablePipeObjectBase):
    """
    表格型模型的PipeLine，注意继承的是TablePipeObjectBase
    """

    def __init__(self, run_to_layer=None, **kwargs):
        """

        :param run_to_layer: 运行到指定的index
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.models = []
        self.run_to_layer = run_to_layer

    def set_dependency(self, model):
        """
        设置model与之前model的依赖关系

        :param model:
        :return:
        """
        # 添加前后依赖
        if len(self.models) > 0:
            model.set_parent_pipe(self.models[-1])
        else:
            model.set_parent_pipe(self.get_parent_pipe())
            return
        # 如果当前添加的model是pipeline或者Parallel,需要递归设置
        need_update_parent_pipes = []
        # 如果当前model也是pipeline
        if issubclass(model.__class__, TablePipeLine) and len(model.models) > 0:
            need_update_parent_pipes.append(model.models[0])
        # 如果当前model是Parallel
        if issubclass(model.__class__, easymlops.table.ensemble.Parallel) and len(model.pipe_objects) > 0:
            need_update_parent_pipes.extend(model.pipe_objects)
        while len(need_update_parent_pipes) > 0:
            pipe_ = need_update_parent_pipes.pop(0)
            pipe_.set_parent_pipe(self.models[-1])
            if issubclass(pipe_.__class__, TablePipeLine) and len(pipe_.models) > 0:
                need_update_parent_pipes.append(pipe_.models[0])
            # 如果当前model是Parallel
            if issubclass(pipe_.__class__, easymlops.table.ensemble.Parallel) and len(pipe_.pipe_objects) > 0:
                need_update_parent_pipes.extend(pipe_.pipe_objects)

    def pipe(self, model):
        """
        添加新pipe

        :param model:
        :return:
        """
        self.set_dependency(model)
        self.models.append(model)
        return self

    def __getitem__(self, index):
        # 切片方式
        if isinstance(index, slice):
            target_models = self.models[index]
            new_pipe_line = self.__class__()
            for model in target_models:
                new_pipe_line.pipe(copy.deepcopy(model))
            return new_pipe_line
        # 读取指定层方式
        elif type(index) == int or type(index) == str or type(index) == list or type(index) == tuple:
            return self._match_index_pipe(match_index=index)
        else:
            raise Exception("{} indices should be int,list,tuple,str or slice".format(self.name))

    def fit(self, x, show_process=False, **kwargs):
        """

        :param x:
        :param show_process: 是否打印训练过程
        :param kwargs:
        :return:
        """
        # 注意:最后一层无需transform
        x_ = copy.deepcopy(x)
        run_idx = range(len(self.models))
        if show_process:
            run_idx = tqdm(run_idx)
        for idx in run_idx:
            model = self.models[idx]
            if show_process:
                print(model.name)
            if idx == len(self.models) - 1:
                model.fit(x_, **kwargs)
                # pipeline的最后一个pipe也要做transform，不过只处理一条数据，为的是调用transform过程中需要保留的一些参数
                model.transform(copy.deepcopy(x_[:1]), **kwargs)
            else:
                x_ = model.fit(x_, **kwargs).transform(x_, **kwargs)
        return self

    def _match_index_pipe(self, match_index) -> TablePipeObjectBase:
        """
        在当前pipeline中寻找匹配index的pipe

        :param match_index:
        :return:
        """
        if type(match_index) == int:
            return self.models[match_index]
        if type(match_index) == tuple:
            match_index = list(match_index)
        if type(match_index) == list:
            # 逐层获取
            pipes_ = []
            pipes_.extend(self.models)
            current_pipe = None
            while len(match_index) > 0:
                index = match_index.pop(0)
                current_pipe = pipes_[index]
                pipes_ = []
                if issubclass(current_pipe.__class__, TablePipeLine):
                    pipes_.extend(current_pipe.models)
                if issubclass(current_pipe.__class__, easymlops.table.ensemble.Parallel):
                    pipes_.extend(current_pipe.pipe_objects)
            return current_pipe
        if type(match_index) == str:
            pipes_ = []
            pipes_.extend(self.models)
            while len(pipes_) > 0:
                ipipe = pipes_.pop(0)
                if ipipe.name == match_index:
                    return ipipe
                if issubclass(ipipe.__class__, easymlops.table.ensemble.Parallel):
                    pipes_.extend(ipipe.pipe_objects)
                if issubclass(ipipe.__class__, TablePipeLine):
                    pipes_.extend(ipipe.models)
        raise Exception(f"can't match the index:{match_index}")

    def transform(self, x, show_process=False, run_to_layer=None, **kwargs):
        """

        :param x:
        :param show_process: 可视化训练过程
        :param run_to_layer: 只运行至指定的index
        :param kwargs:
        :return:
        """
        run_to_layer = self.run_to_layer if run_to_layer is None else run_to_layer
        x_ = copy.deepcopy(x)
        # 处理run_to_layer
        if run_to_layer is not None:
            pipe_ = self._match_index_pipe(run_to_layer)
            run_models = pipe_.get_all_parent_pipes()
            run_models.append(pipe_)
        else:
            run_models = self.models
        for model in tqdm(run_models) if show_process else run_models:
            if show_process:
                print(model.name)
            x_ = model.transform(x_, **kwargs)
        return x_

    def transform_single(self, x, show_process=False, run_to_layer=None, logger=None, prefix="step",
                         log_base_dict: dict_type = None, storage_base_dict: dict_type = None, **kwargs):
        """
        生成部署用于单条数据预测

        :param x:
        :param show_process: 可视化预测过程
        :param run_to_layer: 运行至指定的index
        :param logger: 记录日志，需要提供info和error函数
        :param prefix: 打印日志阶段所用的前缀
        :param log_base_dict: 记录日志的额外补充信息（主要是key）
        :param storage_base_dict: 存储模块调用时的额外补充信息（主要是key）
        :param kwargs:
        :return:
        """
        run_to_layer = self.run_to_layer if run_to_layer is None else run_to_layer
        x_ = copy.deepcopy(x)

        # 处理run_to_layer
        if run_to_layer is not None:
            pipe_ = self._match_index_pipe(run_to_layer)
            run_models = pipe_.get_all_parent_pipes()
            run_models.append(pipe_)
            run_models = enumerate(run_models)
        else:
            run_models = enumerate(self.models)
        for current_layer_deep, model in tqdm(run_models) if show_process else run_models:
            if show_process:
                print(model.name)
            if issubclass(model.__class__, TablePipeLine):
                x_ = model.transform_single(x_, show_process=show_process,
                                            logger=logger,
                                            prefix="{}-{}".format(prefix, current_layer_deep),
                                            log_base_dict=log_base_dict, storage_base_dict=storage_base_dict,
                                            **kwargs)
            else:
                x_ = model.transform_single(x_, logger=logger,
                                            log_base_dict=log_base_dict,
                                            storage_base_dict=storage_base_dict, **kwargs)
                self._save_log(logger=logger, log_base_dict=log_base_dict,
                               step="{}-{}".format(prefix, current_layer_deep), transform=x_, pipe_name=model.name)
        return x_

    @staticmethod
    def _save_log(logger, log_base_dict, step, transform, pipe_name):
        """
        保存日志

        :param logger: 主要提供,logger.info方法，最好异步，不阻塞主程序
        :param log_base_dict: 续保保存基本信息，比如涉及到该条数据的id信息等
        :param step: 所在pipeline的第几步
        :param transform: 该步的输出
        :param pipe_name:pipe模块名称
        :return:
        """
        if log_base_dict is None:
            log_base_dict = dict()
        if logger is not None:
            log_info = {"step": step, "pipe_name": pipe_name, "transform": copy.deepcopy(transform)}
            log_info.update(log_base_dict)
            logger.info(log_info)

    def save(self, path):
        """
        存储模型至path路径

        :param path:
        :return:
        """
        with open(path, "wb") as f:
            pickle.dump([model.get_params() for model in self.models], f)

    def load(self, path):
        """
        从path路径加载模型

        :param path:
        :return:
        """
        with open(path, "rb") as f:
            for i, params in enumerate(pickle.load(f)):
                self.models[i].set_params(params)

    def get_params(self) -> dict:
        """
        获取整个pipeline的参数

        :return:
        """
        params = [model.get_params() for model in self.models]
        return {"params": params}

    def set_params(self, params: dict):
        """
        设置整个pipeline的参数

        :param params:
        :return:
        """
        params = params["params"]
        for i, param in enumerate(params):
            self.models[i].set_params(param)

    def auto_test(self, x, sample=100):
        """
        自动测试接口

        :param x:
        :param sample:
        :return:
        """
        from easymlops.table.callback import check_transform_function_pipeline, check_null_value, check_extreme_value, \
            check_inverse_dtype, check_int_trans_float
        check_transform_function_describe = """
###################################################################
 1.一致性测试和性能测试:check_transform_function                      
###################################################################"""
        print(check_transform_function_describe)
        self.callback(check_transform_function_pipeline, x, sample=sample)
        check_null_value_describe = """
#########################################################################################
 2.空值测试:check_null_value                                                           
 null_values=[None, np.nan, "null", "NULL", "nan", "NaN", "", "none", "None", " "](默认)
########################################################################################"""
        print(check_null_value_describe)
        self.callback(check_null_value, x, sample=sample)
        check_extreme_value_describe = """
############################################################################################################
 3.极端值测试:check_extreme_value                                                                            
 number_extreme_values = [np.inf, 0.0, -1, 1, -1e-7, 1e-7, np.finfo(np.float64).min, np.finfo(np.float64).max](默认)
 category_extreme_values = ["", "null", None, "1.0", "0.0", "-1.0", "-1", "NaN", "None"](默认)                            
###########################################################################################################"""
        print(check_extreme_value_describe)
        self.callback(check_extreme_value, x, sample=sample)
        check_inverse_dtype_describe = """
###############################################################################################################
 4.数据类型反转测试:check_inverse_dtype                                                                          
 category_inverse_values = [0.0, -1, 1, -1e-7, 1e-7, np.finfo(np.float64).min, np.finfo(np.float64).max](默认)
 number_inverse_values = ["", "null", None, "1.0", "0.0", "-1.0", "-1"](默认)                                
#############################################################################################################"""
        print(check_inverse_dtype_describe)
        self.callback(check_inverse_dtype, x, sample=sample)

        check_int_trans_float_describe = """
############################################
 5.int数据转float测试:check_int_trans_float                                                                                            
############################################"""
        print(check_int_trans_float_describe)
        self.callback(check_int_trans_float, x, sample=sample)



from setuptools import find_packages, setup

with open('requirements.txt') as f:
    requirements = f.read()


setup(
    name='easymlops',
    version='0.1.1',
    python_requires='>=3.6',
    author='zhulei227',
    url='https://github.com/zhulei227/EasyMLOps',
    description='MLOps Toolkit In Pipeline',
    packages=find_packages(),
    license='Apache-2.0',
    install_requires=requirements)



from easymlops.table.core import *
import scipy.sparse as sp


class ReduceMemUsage(TablePipeObjectBase):
    """
    功能性：通过修改数据类型减少内存使用量 \n
    """

    def __init__(self, skip_check_transform_type=True, **kwargs):
        super().__init__(skip_check_transform_type=skip_check_transform_type, **kwargs)
        self.type_map_detail = dict()

    @staticmethod
    def extract_dict(s: dict_type, keys: list):
        new_s = dict()
        for key in keys:
            new_s[key] = s[key]
        return new_s

    @staticmethod
    def get_type(ser: series_type):
        col_type = ser.dtype
        col_type_str = str(col_type).lower()
        if col_type != object:
            c_min = ser.min()
            c_max = ser.max()
            if "int" in col_type_str or "bool" in col_type_str:
                int_types = [np.uint8, np.int8, np.uint16, np.int16, np.uint32, np.int32, np.uint64, np.int64]
                for int_type in int_types:
                    if c_min > np.iinfo(int_type).min and c_max < np.iinfo(int_type).max:
                        return int_type, (np.iinfo(int_type).min, np.iinfo(int_type).max)
            elif "float" in col_type_str or "double" in col_type_str:
                float_types = [np.float16, np.float32, np.float64]
                for float_type in float_types:
                    if c_min > np.finfo(float_type).min and c_max < np.finfo(float_type).max:
                        return float_type, (np.finfo(float_type).min, np.finfo(float_type).max)
            else:
                return str, None
        else:
            return str, None

    def udf_fit(self, s: dataframe_type, **kwargs):
        for col in self.input_col_names:
            self.type_map_detail[col] = self.get_type(s[col])
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col, ti in self.type_map_detail.items():
            tp, ran = ti
            try:
                if ran is not None:
                    s[col] = tp(np.clip(s[col], ran[0], ran[1]))
                else:
                    s[col] = s[col].astype(str)
            except Exception as e:
                print(e)
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col, ti in self.type_map_detail.items():
            tp, ran = ti
            try:
                if ran is not None:
                    s[col] = tp(np.clip(s[col], ran[0], ran[1]))
                else:
                    s[col] = str(s[col])
            except Exception as e:
                print(e)
        return s

    def udf_get_params(self) -> dict_type:
        return {"type_map_detail": self.type_map_detail}

    def udf_set_params(self, params: dict):
        self.type_map_detail = params["type_map_detail"]


class Dense2Sparse(ReduceMemUsage):
    """
    功能性：通过将稠密矩阵压缩为稀疏矩阵减少内存使用； \n
    注意：1）矩阵中0比较多时效果显著；2）注意后续pipe object要支持csr结构的稀疏矩阵
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        if not hasattr(s, "sparse"):
            s = pd.DataFrame.sparse.from_spmatrix(data=sp.csr_matrix(s), columns=self.input_col_names, index=s.index)
        for col, ti in self.type_map_detail.items():
            tp, ran = ti
            try:
                if ran is not None:
                    s[col] = s[col].astype(tp)
                else:
                    s[col] = s[col].astype(str)
            except Exception as e:
                print(e)
        return s

    def udf_get_params(self) -> dict_type:
        return {}



import time
import datetime
from easymlops.table.core import *
import threading
import sqlite3

lock = threading.Lock()


class LocalStorage(TablePipeObjectBase, threading.Thread):
    """
    存储模块，主要用于在transform_single阶段存储数据,其余过程均透传，这里默认对key创建了索引
    """

    def __init__(self, cols=None, db_name="local.db", table_name="temp", **kwargs):
        super().__init__(**kwargs)
        self.cols = cols
        if self.cols is None:
            self.cols = ["value"]
        self.table_columns = ["storage_key", "storage_transform_time"] + self.cols
        self.db_name = db_name
        self.table_name = table_name
        self.cache_data = []
        self._conn, self._cur = self.create_connect(self.db_name, self.table_name)
        self.stop_flag = False
        self.start()

    def reconnect(self):
        self._conn, self._cur = self.create_connect(self.db_name, self.table_name)

    def create_connect(self, db_name, table_name):
        _conn = None
        _cur = None
        # 创建链接
        try:
            _conn = sqlite3.connect(db_name, timeout=1, check_same_thread=False)
            _cur = _conn.cursor()
            columns_str = ",".join([item + " varchar" for item in self.table_columns])
            sql = f"create table if not exists {table_name} " \
                  f"({columns_str})"
            _cur.execute(sql)
        except Exception as e:
            print(f"connect to {db_name}:{table_name} exception:{e}")
        # 创建索引:storage_key
        if _conn is not None and _cur is not None:
            try:
                sql = f"create index if not exists index_{table_name}_storage_key on {table_name}(storage_key)"
                _cur.execute(sql)
            except Exception as e:
                print(f"create index storage_key exception:{e}")
        # 创建索引:storage_transform_time
        if _conn is not None and _cur is not None:
            try:
                sql = f"create index if not exists index_{table_name}_storage_transform_time " \
                      f"on {table_name}(storage_transform_time)"
                _cur.execute(sql)
            except Exception as e:
                print(f"create index storage_transform_time exception:{e}")
        return _conn, _cur

    def fit(self, s: dataframe_type, **kwargs):
        return self

    def transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        return s

    def udf_get_params(self) -> dict_type:
        return {"db_name": self.db_name, "table_name": self.table_name, "table_columns": self.table_columns}

    def udf_set_params(self, params: dict_type):
        self.db_name = params["db_name"]
        self.table_name = params["table_name"]
        self.reconnect()
        self.table_columns = params["table_columns"]

    def update_connect(self, db_name, table_name):
        self.db_name = db_name
        self.table_name = table_name
        self._conn, self._cur = self.create_connect(self.db_name, self.table_name)

    def transform_single(self, s: dict_type, storage_base_dict: dict_type = None, **kwargs) -> dict_type:
        s_ = copy.deepcopy(s)
        s_["storage_key"] = "default" if storage_base_dict is None \
            else storage_base_dict.get("key", "default")
        s_["storage_transform_time"] = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        register_data = []
        for col in self.table_columns:
            register_data.append(str(s_.get(col, "")))
        self.cache_data.append(tuple(register_data))
        return s

    def select_key(self, key="default", limit=10):
        sql = f"select * from {self.table_name} where storage_key='{key}' limit {limit}"
        try:
            lock.acquire(True)
            self._cur.execute(sql)
            columns = [tp[0] for tp in self._cur.description]
            return pd.DataFrame(self._cur.fetchmany(size=limit), columns=columns)
        except Exception as e:
            print(f"[select] data exception:{e},try reconnect db")
            self._conn, self._cur = self.create_connect(self.db_name, self.table_name)
        finally:
            lock.release()

    def where(self, where_sql="", limit=10):
        sql = f"select * from {self.table_name} where {where_sql} limit {limit}"
        try:
            lock.acquire(True)
            self._cur.execute(sql)
            columns = [tp[0] for tp in self._cur.description]
            return pd.DataFrame(self._cur.fetchmany(size=limit), columns=columns)
        except Exception as e:
            print(f"[where] data exception:{e},try reconnect db")
            self._conn, self._cur = self.create_connect(self.db_name, self.table_name)
        finally:
            lock.release()

    def group_agg_where(self, group_by="", agg_sql="", where_sql="", limit=10):
        if type(group_by) == list:
            group_by = ",".join(group_by)
        if group_by is not None and len(group_by) > 0:
            group_by = " group by " + group_by
        if where_sql is not None and len(where_sql) > 0:
            where_sql = " where " + where_sql
        sql = f"select {agg_sql} from {self.table_name} {where_sql} {group_by} limit {limit}"
        try:
            lock.acquire(True)
            self._cur.execute(sql)
            columns = [tp[0] for tp in self._cur.description]
            return pd.DataFrame(self._cur.fetchall(), columns=columns)
        except Exception as e:
            print(f"[group_agg_where] data exception:{e},try reconnect db")
            self._conn, self._cur = self.create_connect(self.db_name, self.table_name)
        finally:
            lock.release()

    def sql(self, sql=""):
        try:
            lock.acquire(True)
            self._cur.execute(sql)
            columns = [tp[0] for tp in self._cur.description]
            return pd.DataFrame(self._cur.fetchall(), columns=columns)
        except Exception as e:
            print(f"[where] data exception:{e},try reconnect db")
            self._conn, self._cur = self.create_connect(self.db_name, self.table_name)
        finally:
            lock.release()

    def set_stop(self):
        self.stop_flag = True

    def run(self):
        while True:
            if self.stop_flag:
                break
            time.sleep(0.001)
            if len(self.cache_data) > 0:
                saved_cache_data = self.cache_data
                self.cache_data = []
                try:
                    lock.acquire(True)
                    cols = ",".join(["?"] * len(self.table_columns))
                    sql = f"INSERT INTO {self.table_name} VALUES ({cols})"
                    self._cur.executemany(sql, saved_cache_data)
                    self._conn.commit()
                except Exception as e:
                    print(f"[insert] data exception:{e},try reconnect db")
                    self._conn, self._cur = self.create_connect(self.db_name, self.table_name)
                finally:
                    lock.release()
